mod op_codes;

use op_codes::*;
use std::alloc::{alloc_zeroed, Layout};

pub fn run(code: &Vec<u8>, registers: &mut Vec<u64>, register_types: &mut Vec<RegisterType>) {
    let mut i = 0;
    while i < code.len() {
        match code[i] {
            MOV_INT => {
                let target_register = code[i + 1] as usize;
                let operand: u32 = code[i + 2] as u32
                    + ((code[i + 3] as u32) << 8)
                    + ((code[i + 4] as u32) << 16)
                    + ((code[i + 5] as u32) << 24);
                registers[target_register] = operand as u64;
                register_types[target_register] = RegisterType::Int;
                i += 6
            }
            MOV_LONG => {
                let target_register = code[i + 1] as usize;
                let operand: u64 = code[i + 2] as u64
                    + ((code[i + 3] as u64) << 8)
                    + ((code[i + 4] as u64) << 16)
                    + ((code[i + 5] as u64) << 24)
                    + ((code[i + 6] as u64) << 32)
                    + ((code[i + 7] as u64) << 40)
                    + ((code[i + 8] as u64) << 48)
                    + ((code[i + 9] as u64) << 56);
                registers[target_register] = operand;
                register_types[target_register] = RegisterType::Long;
                i += 10
            }
            MOV_DOUBLE => { /* todo */ }
            MOV_REGISTER => {
                let target_register = code[i + 1] as usize;
                let source_register = code[i + 2] as usize;
                registers[target_register] = registers[source_register];
                register_types[target_register] = register_types[source_register].clone();
                i += 3;
            }
            ALLOC => {
                let target_register = code[i + 1] as usize;
                let n_bytes_to_allocate = code[i + 2] as u32
                    + ((code[i + 3] as u32) << 8)
                    + ((code[i + 4] as u32) << 16)
                    + ((code[i + 5] as u32) << 24);
                let layout = Layout::from_size_align(n_bytes_to_allocate as usize, 4).unwrap();
                let allocated = unsafe { alloc_zeroed(layout) };
                registers[target_register] = allocated as u64;
                register_types[target_register] = RegisterType::Pointer;
                i += 6;
            }
            MOV_INT_TO => {
                let target_register = code[i + 1] as usize;
                if register_types[target_register] != RegisterType::Pointer {
                    panic!("target_register {} is not a Pointer", target_register);
                }
                let offset = convert_to_u32(&code[(i + 2)..(i + 6)]) as isize;
                let target = registers[target_register] as *mut u8;
                unsafe {
                    target.offset(offset).write(code[i + 6]);
                    target.offset(offset + 1).write(code[i + 7]);
                    target.offset(offset + 2).write(code[i + 8]);
                    target.offset(offset + 3).write(code[i + 9]);
                }
                i += 10;
            }
            _ => panic!("Invalid code instruction"),
        }
    }
}

fn convert_to_u32(bytes: &[u8]) -> u32 {
    bytes[0] as u32 | (bytes[1] as u32) << 8 | (bytes[2] as u32) << 16 | (bytes[3] as u32) << 24
}

#[cfg(test)]
mod tests {
    use super::*;

    fn init_registers(n_registers: usize) -> (Vec<u64>, Vec<RegisterType>) {
        (vec![0; n_registers], vec![RegisterType::Int; n_registers])
    }

    #[test]
    fn mov_1_as_int() {
        let mut code = Vec::new();
        add_mov_int(&mut code, 0, 1);
        let (mut registers, mut register_types) = init_registers(1);
        run(&code, &mut registers, &mut register_types);
        assert_eq!(1, registers[0]);
    }

    #[test]
    fn move_65535_as_int() {
        let mut code = Vec::new();
        add_mov_int(&mut code, 0, 0xffff);
        let (mut registers, mut register_types) = init_registers(1);
        run(&code, &mut registers, &mut register_types);
        assert_eq!(0xffff, registers[0]);
    }

    #[test]
    fn move_int_max_as_int() {
        let mut code = Vec::new();
        add_mov_int(&mut code, 0, 0xffff_ffff);
        let (mut registers, mut register_types) = init_registers(1);
        run(&code, &mut registers, &mut register_types);
        assert_eq!(0xffff_ffff, registers[0]);
    }

    #[test]
    fn move_register() {
        let mut code = Vec::new();
        add_mov_register(&mut code, 0, 1);
        let (mut registers, mut register_types) = init_registers(2);
        registers[1] = 1;
        run(&code, &mut registers, &mut register_types);
        assert_eq!(registers[0], 1);
    }

    #[test]
    fn mov_int_to_register_as_address() {
        let mut code = Vec::new();
        add_alloc(&mut code, 0, 4);
        add_mov_int_to(&mut code, 0, 0, 0xff);
        let mut registers = vec![0; 16];
        let mut register_types = vec![RegisterType::Int; 16];
        run(&code, &mut registers, &mut register_types);
        let target = registers[0] as *mut u8;
        unsafe {
            assert_eq!(0xff, *target);
        }
    }
}
